// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 

// Copyright 2001 Liam Girdwood

#include <iostream>

#include "astro_object.hh"
#include "object_cache.hh"

// instance pointer
object_cache::object_cache * object_cache::m_cache = 0;

// astro object cache

object_cache::object_cache()
{
	int i,j,k;
	// clear sectors ra first
	for (i = 0; i < SECT_RA; i++)
	{
		//now dec
		for (j = 0; j < SECT_DEC; j++)
		{
			// now object types
			for (k=0; k < SECT_TYPE; k++)
			{
				m_sector_head[i][j][k] = 0;
				m_sector_size[i][j][k] = 0;
				m_sector_type[i][j][k] = -1;
			}
		}
	}
	m_max_cache_size = CACHE_SIZE;
}

object_cache::~object_cache()
{
	int i,j;
	
	// clear sectors ra first
	for (i = 0; i < SECT_RA; i++)
	{
		//now dec
		for (j = 0; j < SECT_DEC; j++)
		{
			flush_sector (i,j,0);
		}
	}
}

astro_object * object_cache::get_sector_head (int ra, int dec, int type)
{
	int i;
	
	if (ra >= SECT_RA || dec >= SECT_DEC)
		return (0);
		
	for (i = 0; i < SECT_TYPE; i++)
	{
		if (m_sector_type[ra][dec][i] == type)
		{
			return (m_sector_head[ra][dec][i]);
		}
	}
	return (0);
}


astro_object * object_cache::get_sector_tail (int ra, int dec, int type)
{
	int i;
	
	if (ra >= SECT_RA || dec >= SECT_DEC)
		return (0);
			
	for (i = 0; i < SECT_TYPE; i++)
	{
		if (m_sector_type[ra][dec][i] == type) 
		{
			return (m_sector_tail[ra][dec][i]);
		}
	}
	return (0);
}

int object_cache::flush_sector (int ra, int dec, int type)
{
	int i;
	astro_object * head = 0;
	int bytes = 0;
	
	if (ra >= SECT_RA || dec >= SECT_DEC)
		return (0);
					
	if (type == ASTRO_ALL)
	{
		// flush all types
		for (i = 0; i < SECT_TYPE; i++)
		{
			if ((head = m_sector_head[ra][dec][i]) != 0)
			{
				bytes += delete_list(head);
				m_sector_head[ra][dec][i] = 0;
				m_sector_tail[ra][dec][i] = 0;
				m_sector_size[ra][dec][i] = 0;
				m_sector_type[ra][dec][i] = -1;
			}
		}
		m_cache_size -= bytes;
	}
	else
	{
		//flush type type
		for (i = 0; i < SECT_TYPE; i++)
		{
			if (m_sector_type[ra][dec][i] == type)
			{
				head = m_sector_head[ra][dec][i];
				break;
			}
		}
		if (head)
		{
			bytes = delete_list (head);
			m_sector_head[ra][dec][i] = 0;
			m_sector_tail[ra][dec][i] = 0;
			m_sector_size[ra][dec][i] = 0;
			m_cache_size -= bytes;
		}
		else
			return (0);
	}
	return (bytes);
}

int object_cache::delete_list (astro_object * head)
{
	astro_object * next;
	int bytes = 0;
	int size = head->get_size();
	
	do 
	{
		next = head->tail;
		delete (head);
		head = next;
		bytes += size;
	}
	while (next);
	
	return (bytes);
}
		

int object_cache::get_sector_size (int ra, int dec, int type)
{
	int i;
	
	if (ra >= SECT_RA || dec >= SECT_DEC)
		return (0);
	
	for (i = 0; i < SECT_TYPE; i++)
	{
		if (m_sector_type[ra][dec][i] == type)
		{
			return (m_sector_size[ra][dec][i]);
		}
	}
	return (0);
}

int object_cache::add_object (astro_object * object)
{
	int ra, dec;
	int i;
	int free_slot = 0;
	bool found = 0;
	int size = object->get_size();
	astro_object * tail;
	
	// check against size
	if (size + m_cache_size > m_max_cache_size)
		return (0);
	
	ra = (int) object->astro.RA;
	dec = (int) ((object->astro.DEC + 90) / 10);
	
	if (ra >= SECT_RA || ra < 0 || dec > SECT_DEC || dec < 0 )
	{
		std::cerr << "Cache add: got garbage ra " << ra << " dec " << dec << endl;
		return 0;
	}
	
	// find existing type  slot
	for (i = 0; i < SECT_TYPE; i++)
	{
		if (m_sector_type[ra][dec][i] == object->obj_type)
		{
			found = 1;
			break;
		}
	}
	
	// if type is new then find a free slot
	if (!found)
	{
		for (free_slot = 0; free_slot < SECT_TYPE; free_slot++)
		{
			if (m_sector_type[ra][dec][free_slot] < 0)
				break;
		}
	}
	
	// check for existing or new slot
	if (!found && free_slot == SECT_TYPE)
	{
		return (0);  // not enough room
	}

	if (found)
	{
		// add object to end of list
		tail = m_sector_tail[ra][dec][i];
		tail->tail = object;
		m_sector_tail[ra][dec][i] = object;
		m_sector_size[ra][dec][i] += size;
		m_cache_size += size;
	}
	else
	{
		// make object head of new list
		m_sector_head[ra][dec][free_slot] = object;
		m_sector_tail[ra][dec][free_slot] = object;
		m_sector_size[ra][dec][free_slot] = size;
		m_sector_type[ra][dec][free_slot] = object->obj_type;
		m_cache_size += size;
	}
	
}

int object_cache::get_cache_size ()
{
	return (m_cache_size);
}

void object_cache::set_cache_size (int bytes)
{
	m_max_cache_size = bytes;
}

int object_cache::get_cache_free ()
{
	return (m_max_cache_size - m_cache_size);
}

object_cache * object_cache::instance ()
{
	if (!m_cache)
	{
		m_cache = new object_cache ();
	}
	return m_cache;
}

void object_cache::release ()
{
//	delete m_cache;
//	m_cache = 0;
}
